@@ -5,34 +5,42 @@ Big thanks goes out to Fred Clift <fred@clift.org> who build the first
 version of this code: https://github.com/minektur/chromecast-python-poc.
 Without him this would not have been possible.
 """
+
 # pylint: disable=too-many-lines
+from __future__ import annotations
 
 import abc
+from dataclasses import dataclass
 import errno
 import json
 import logging
 import select
 import socket
 import ssl
-import sys
 import threading
 import time
-from collections import defaultdict, namedtuple
+from collections import defaultdict
 from struct import pack, unpack
 
-from . import cast_channel_pb2
-from .controllers import BaseController
+import zeroconf
+
+from .controllers import CallbackType, BaseController
 from .controllers.media import MediaController
-from .controllers.receiver import ReceiverController
+from .controllers.receiver import CastStatus, CastStatusListener, ReceiverController
 from .const import MESSAGE_TYPE, REQUEST_ID, SESSION_ID
 from .dial import get_host_from_service
 from .error import (
     ChromecastConnectionError,
+    ControllerNotRegistered,
     UnsupportedNamespace,
     NotConnected,
     PyChromecastStopped,
 )
 
+# pylint: disable-next=no-name-in-module
+from .generated.cast_channel_pb2 import CastMessage
+from .models import HostServiceInfo, MDNSServiceInfo
+
 NS_CONNECTION = "urn:x-cast:com.google.cast.tp.connection"
 NS_HEARTBEAT = "urn:x-cast:com.google.cast.tp.heartbeat"
 
@@ -63,15 +71,15 @@ HB_PING_TIME = 10
 HB_PONG_TIME = 10
 POLL_TIME_BLOCKING = 5.0
 POLL_TIME_NON_BLOCKING = 0.01
-TIMEOUT_TIME = 30
-RETRY_TIME = 5
+TIMEOUT_TIME = 30.0
+RETRY_TIME = 5.0
 
 
 class InterruptLoop(Exception):
     """The chromecast has been manually stopped."""
 
 
-def _dict_from_message_payload(message):
+def _dict_from_message_payload(message: CastMessage) -> dict:
     """Parses a PB2 message as a JSON dict."""
     try:
         data = json.loads(message.payload_utf8)
@@ -94,7 +102,10 @@ def _dict_from_message_payload(message):
         return {}
 
 
-def _message_to_string(message, data=None):
+def _message_to_string(
+    message: CastMessage,
+    data: dict | None = None,
+) -> str:
     """Gives a string representation of a PB2 message."""
     if data is None:
         data = _dict_from_message_payload(message)
@@ -105,43 +116,33 @@ def _message_to_string(message, data=None):
     )
 
 
-if sys.version_info >= (3, 0):
+@dataclass(frozen=True)
+class NetworkAddress:
+    """Network address container."""
 
-    def _json_to_payload(data):
-        """Encodes a python value into JSON format."""
-        return json.dumps(data, ensure_ascii=False).encode("utf8")
-
-else:
-
-    def _json_to_payload(data):
-        """Encodes a python value into JSON format."""
-        return json.dumps(data, ensure_ascii=False)
+    address: str
+    port: int | None
 
 
-def _is_ssl_timeout(exc):
-    """Returns True if the exception is for an SSL timeout"""
-    return exc.message in (
-        "The handshake operation timed out",
-        "The write operation timed out",
-        "The read operation timed out",
-    )
+@dataclass(frozen=True)
+class ConnectionStatus:
+    """Connection status container."""
 
-
-NetworkAddress = namedtuple("NetworkAddress", ["address", "port"])
-
-ConnectionStatus = namedtuple("ConnectionStatus", ["status", "address"])
+    status: str
+    address: NetworkAddress | None
+    service: HostServiceInfo | MDNSServiceInfo | None
 
 
 class ConnectionStatusListener(abc.ABC):
     """Listener for receiving connection status events."""
 
     @abc.abstractmethod
-    def new_connection_status(self, status: ConnectionStatus):
+    def new_connection_status(self, status: ConnectionStatus) -> None:
         """Updated connection status."""
 
 
 # pylint: disable-next=too-many-instance-attributes
-class SocketClient(threading.Thread):
+class SocketClient(threading.Thread, CastStatusListener):
     """
     Class to interact with a Chromecast through a socket.
 
@@ -173,13 +174,13 @@ class SocketClient(threading.Thread):
     def __init__(
         self,
         *,
-        cast_type,
-        tries,
-        timeout,
-        retry_wait,
-        services,
-        zconf,
-    ):
+        cast_type: str,
+        tries: int | None,
+        timeout: float | None,
+        retry_wait: float | None,
+        services: set[HostServiceInfo | MDNSServiceInfo],
+        zconf: zeroconf.Zeroconf | None,
+    ) -> None:
         super().__init__()
 
         self.daemon = True
@@ -189,7 +190,7 @@ class SocketClient(threading.Thread):
         self._force_recon = False
 
         self.cast_type = cast_type
-        self.fn = None  # pylint:disable=invalid-name
+        self.fn: str | None = None  # pylint:disable=invalid-name
         self.tries = tries
         self.timeout = timeout or TIMEOUT_TIME
         self.retry_wait = retry_wait or RETRY_TIME
@@ -204,21 +205,20 @@ class SocketClient(threading.Thread):
         # socketpair used to interrupt the worker thread
         self.socketpair = socket.socketpair()
 
-        self.app_namespaces = []
-        self.destination_id = None
-        self.session_id = None
+        self.app_namespaces: list[str] = []
+        self.destination_id: str | None = None
+        self.session_id: str | None = None
         self._request_id = 0
-        # dict mapping requestId to callback functions
-        self._request_callbacks = {}
-        self._open_channels = []
+        self._request_callbacks: dict[int, CallbackType] = {}
+        self._open_channels: list[str] = []
 
         self.connecting = True
         self.first_connection = True
-        self.socket = None
+        self.socket: socket.socket | ssl.SSLSocket | None = None
 
         # dict mapping namespace on Controller objects
-        self._handlers = defaultdict(set)
-        self._connection_listeners = []
+        self._handlers: dict[str, set[BaseController]] = defaultdict(set)
+        self._connection_listeners: list[ConnectionStatusListener] = []
 
         self.receiver_controller = ReceiverController(cast_type)
         self.media_controller = MediaController()
@@ -231,9 +231,9 @@ class SocketClient(threading.Thread):
 
         self.receiver_controller.register_status_listener(self)
 
-    def initialize_connection(
+    def initialize_connection(  # pylint:disable=too-many-statements, too-many-branches
         self,
-    ):  # pylint:disable=too-many-statements, too-many-branches
+    ) -> None:
         """Initialize a socket to a Chromecast, retrying as necessary."""
         tries = self.tries
 
@@ -241,6 +241,10 @@ class SocketClient(threading.Thread):
             self.socket.close()
             self.socket = None
 
+        # Make sure nobody is blocking.
+        for callback_function in self._request_callbacks.values():
+            callback_function(False, None)
+
         self.app_namespaces = []
         self.destination_id = None
         self.session_id = None
@@ -252,9 +256,12 @@ class SocketClient(threading.Thread):
         retry_log_fun = self.logger.error
 
         # Dict keeping track of individual retry delay for each named service
-        retries = {}
+        retries: dict[HostServiceInfo | MDNSServiceInfo, dict[str, float]] = {}
 
-        def mdns_backoff(service, retry):
+        def mdns_backoff(
+            service: HostServiceInfo | MDNSServiceInfo,
+            retry: dict[str, float],
+        ) -> None:
             """Exponentional backoff for service name mdns lookups."""
             now = time.time()
             retry["next_retry"] = now + retry["delay"]
@@ -280,7 +287,8 @@ class SocketClient(threading.Thread):
                     continue
                 try:
                     if self.socket is not None:
-                        self.socket.close()
+                        # If we retry connecting, we need to clean up the socket again
+                        self.socket.close()  # type: ignore[unreachable]
                         self.socket = None
 
                     self.socket = new_socket()
@@ -289,6 +297,7 @@ class SocketClient(threading.Thread):
                         ConnectionStatus(
                             CONNECTION_STATUS_CONNECTING,
                             NetworkAddress(self.host, self.port),
+                            None,
                         )
                     )
                     # Resolve the service name.
@@ -300,7 +309,8 @@ class SocketClient(threading.Thread):
                     if host and port:
                         if service_info:
                             try:
-                                self.fn = service_info.properties[b"fn"].decode("utf-8")
+                                # Mypy does not understand that we catch errors, ignore it
+                                self.fn = service_info.properties[b"fn"].decode("utf-8")  # type: ignore[union-attr]
                             except (AttributeError, KeyError, UnicodeError):
                                 pass
                         self.logger.debug(
@@ -325,7 +335,8 @@ class SocketClient(threading.Thread):
                         self._report_connection_status(
                             ConnectionStatus(
                                 CONNECTION_STATUS_FAILED_RESOLVE,
-                                NetworkAddress(service, None),
+                                None,
+                                service,
                             )
                         )
                         mdns_backoff(service, retry)
@@ -342,7 +353,9 @@ class SocketClient(threading.Thread):
                         self.port,
                     )
                     self.socket.connect((self.host, self.port))
-                    context = ssl.SSLContext()
+                    context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
+                    context.check_hostname = False
+                    context.verify_mode = ssl.CERT_NONE
                     self.socket = context.wrap_socket(self.socket)
                     self.connecting = False
                     self._force_recon = False
@@ -350,6 +363,7 @@ class SocketClient(threading.Thread):
                         ConnectionStatus(
                             CONNECTION_STATUS_CONNECTED,
                             NetworkAddress(self.host, self.port),
+                            None,
                         )
                     )
                     self.receiver_controller.update_status()
@@ -392,26 +406,18 @@ class SocketClient(threading.Thread):
                         ConnectionStatus(
                             CONNECTION_STATUS_FAILED,
                             NetworkAddress(self.host, self.port),
+                            None,
                         )
                     )
-                    if service is not None:
-                        retry_log_fun(
-                            "[%s(%s):%s] Failed to connect to service %s, retrying in %.1fs",
-                            self.fn or "",
-                            self.host,
-                            self.port,
-                            service,
-                            retry["delay"],
-                        )
-                        mdns_backoff(service, retry)
-                    else:
-                        retry_log_fun(
-                            "[%s(%s):%s] Failed to connect, retrying in %.1fs",
-                            self.fn or "",
-                            self.host,
-                            self.port,
-                            self.retry_wait,
-                        )
+                    retry_log_fun(
+                        "[%s(%s):%s] Failed to connect to service %s, retrying in %.1fs",
+                        self.fn or "",
+                        self.host,
+                        self.port,
+                        service,
+                        retry["delay"],
+                    )
+                    mdns_backoff(service, retry)
                     retry_log_fun = self.logger.debug
 
             # Only sleep if we have another retry remaining
@@ -438,22 +444,7 @@ class SocketClient(threading.Thread):
         )
         raise ChromecastConnectionError("Failed to connect")
 
-    def connect(self):
-        """Connect socket connection to Chromecast device.
-
-        Must only be called if the worker thread will not be started.
-        """
-        try:
-            self.initialize_connection()
-        except ChromecastConnectionError:
-            self._report_connection_status(
-                ConnectionStatus(
-                    CONNECTION_STATUS_DISCONNECTED, NetworkAddress(self.host, self.port)
-                )
-            )
-            return
-
-    def disconnect(self):
+    def disconnect(self) -> None:
         """Disconnect socket connection to Chromecast device"""
         self.stop.set()
         try:
@@ -463,13 +454,13 @@ class SocketClient(threading.Thread):
             # The socketpair may already be closed during shutdown, ignore it
             pass
 
-    def register_handler(self, handler: BaseController):
+    def register_handler(self, handler: BaseController) -> None:
         """Register a new namespace handler."""
         self._handlers[handler.namespace].add(handler)
 
         handler.registered(self)
 
-    def unregister_handler(self, handler: BaseController):
+    def unregister_handler(self, handler: BaseController) -> None:
         """Register a new namespace handler."""
         if (
             handler.namespace in self._handlers
@@ -479,34 +470,34 @@ class SocketClient(threading.Thread):
 
         handler.unregistered()
 
-    def new_cast_status(self, cast_status):
+    def new_cast_status(self, status: CastStatus) -> None:
         """Called when a new cast status has been received."""
-        new_channel = self.destination_id != cast_status.transport_id
+        new_channel = self.destination_id != status.transport_id
 
-        if new_channel:
+        if new_channel and self.destination_id is not None:
             self.disconnect_channel(self.destination_id)
 
-        self.app_namespaces = cast_status.namespaces
-        self.destination_id = cast_status.transport_id
-        self.session_id = cast_status.session_id
+        self.app_namespaces = status.namespaces
+        self.destination_id = status.transport_id
+        self.session_id = status.session_id
 
-        if new_channel:
+        if new_channel and self.destination_id is not None:
             # If any of the namespaces of the new app are supported
             # we will automatically connect to it to receive updates
             for namespace in self.app_namespaces:
                 if namespace in self._handlers:
                     self._ensure_channel_connected(self.destination_id)
-                    for handler in self._handlers[namespace]:
+                    for handler in set(self._handlers[namespace]):
                         handler.channel_connected()
 
-    def _gen_request_id(self):
+    def _gen_request_id(self) -> int:
         """Generates a unique request id."""
         self._request_id += 1
 
         return self._request_id
 
     @property
-    def is_connected(self):
+    def is_connected(self) -> bool:
         """
         Returns True if the client is connected, False if it is stopped
         (or trying to connect).
@@ -514,21 +505,23 @@ class SocketClient(threading.Thread):
         return not self.connecting
 
     @property
-    def is_stopped(self):
+    def is_stopped(self) -> bool:
         """
         Returns True if the connection has been stopped, False if it is
         running.
         """
         return self.stop.is_set()
 
-    def run(self):
+    def run(self) -> None:
         """Connect to the cast and start polling the socket."""
         try:
             self.initialize_connection()
         except ChromecastConnectionError:
             self._report_connection_status(
                 ConnectionStatus(
-                    CONNECTION_STATUS_DISCONNECTED, NetworkAddress(self.host, self.port)
+                    CONNECTION_STATUS_DISCONNECTED,
+                    NetworkAddress(self.host, self.port),
+                    None,
                 )
             )
             return
@@ -537,7 +530,7 @@ class SocketClient(threading.Thread):
         self.logger.debug("Thread started...")
         while not self.stop.is_set():
             try:
-                if self.run_once(timeout=POLL_TIME_BLOCKING) == 1:
+                if self._run_once(timeout=POLL_TIME_BLOCKING) == 1:
                     break
             except Exception:  # pylint: disable=broad-except
                 self._force_recon = True
@@ -552,11 +545,8 @@ class SocketClient(threading.Thread):
         # Clean up
         self._cleanup()
 
-    def run_once(self, timeout=POLL_TIME_NON_BLOCKING):
-        """
-        Use run_once() in your own main loop after you
-        receive something on the socket (get_socket()).
-        """
+    def _run_once(self, timeout: float = POLL_TIME_NON_BLOCKING) -> int:
+        """Receive from the socket and handle data."""
         # pylint: disable=too-many-branches, too-many-statements, too-many-return-statements
 
         try:
@@ -565,6 +555,9 @@ class SocketClient(threading.Thread):
         except ChromecastConnectionError:
             return 1
 
+        # A connection has been established at this point by self._check_connection
+        assert self.socket is not None
+
         # poll the socket, as well as the socketpair to allow us to be interrupted
         rlist = [self.socket, self.socketpair[0]]
         try:
@@ -591,8 +584,8 @@ class SocketClient(threading.Thread):
             self._force_recon = True
             return 0
 
-        # read messages from chromecast
-        message = data = None
+        # read message from chromecast
+        message = None
         if self.socket in can_read and not self._force_recon:
             try:
                 message = self._read_message()
@@ -645,18 +638,11 @@ class SocketClient(threading.Thread):
         self._route_message(message, data)
 
         if REQUEST_ID in data and data[REQUEST_ID] in self._request_callbacks:
-            self._request_callbacks.pop(data[REQUEST_ID], None)(data)
+            self._request_callbacks.pop(data[REQUEST_ID])(True, data)
 
         return 0
 
-    def get_socket(self):
-        """
-        Returns the socket of the connection to use it in you own
-        main loop.
-        """
-        return self.socket
-
-    def _check_connection(self):
+    def _check_connection(self) -> bool:
         """
         Checks if the connection is active, and if not reconnect
 
@@ -689,7 +675,7 @@ class SocketClient(threading.Thread):
                 self.disconnect_channel(channel)
             self._report_connection_status(
                 ConnectionStatus(
-                    CONNECTION_STATUS_LOST, NetworkAddress(self.host, self.port)
+                    CONNECTION_STATUS_LOST, NetworkAddress(self.host, self.port), None
                 )
             )
             try:
@@ -699,7 +685,7 @@ class SocketClient(threading.Thread):
             return False
         return True
 
-    def _route_message(self, message, data: dict):
+    def _route_message(self, message: CastMessage, data: dict) -> None:
         """Route message to any handlers on the message namespace"""
         # route message to handlers
         if message.namespace in self._handlers:
@@ -714,7 +700,7 @@ class SocketClient(threading.Thread):
                 )
 
             # message handlers
-            for handler in self._handlers[message.namespace]:
+            for handler in set(self._handlers[message.namespace]):
                 try:
                     handled = handler.receive_message(message, data)
 
@@ -749,7 +735,7 @@ class SocketClient(threading.Thread):
                 _message_to_string(message, data),
             )
 
-    def _cleanup(self):
+    def _cleanup(self) -> None:
         """Cleanup open channels and handlers"""
         for channel in self._open_channels:
             try:
@@ -757,8 +743,8 @@ class SocketClient(threading.Thread):
             except Exception:  # pylint: disable=broad-except
                 pass
 
-        for namespace in self._handlers.values():
-            for handler in namespace:
+        for handlers in self._handlers.values():
+            for handler in set(handlers):
                 try:
                     handler.tear_down()
                 except Exception:  # pylint: disable=broad-except
@@ -773,7 +759,9 @@ class SocketClient(threading.Thread):
                 )
         self._report_connection_status(
             ConnectionStatus(
-                CONNECTION_STATUS_DISCONNECTED, NetworkAddress(self.host, self.port)
+                CONNECTION_STATUS_DISCONNECTED,
+                NetworkAddress(self.host, self.port),
+                None,
             )
         )
 
@@ -782,7 +770,7 @@ class SocketClient(threading.Thread):
 
         self.connecting = True
 
-    def _report_connection_status(self, status):
+    def _report_connection_status(self, status: ConnectionStatus) -> None:
         """Report a change in the connection status to any listeners"""
         for listener in self._connection_listeners:
             try:
@@ -804,8 +792,11 @@ class SocketClient(threading.Thread):
                     self.port,
                 )
 
-    def _read_bytes_from_socket(self, msglen):
+    def _read_bytes_from_socket(self, msglen: int) -> bytes:
         """Read bytes from the socket."""
+        # It is a programming error if this is called when we don't have a socket
+        assert self.socket is not None
+
         chunks = []
         bytes_recd = 0
         while bytes_recd < msglen:
@@ -817,7 +808,7 @@ class SocketClient(threading.Thread):
                     raise socket.error("socket connection broken")
                 chunks.append(chunk)
                 bytes_recd += len(chunk)
-            except socket.timeout:
+            except TimeoutError:
                 self.logger.debug(
                     "[%s(%s):%s] timeout in : _read_bytes_from_socket",
                     self.fn or "",
@@ -825,21 +816,9 @@ class SocketClient(threading.Thread):
                     self.port,
                 )
                 continue
-            except ssl.SSLError as exc:
-                # Support older ssl implementations which does not raise
-                # socket.timeout on timeouts
-                if _is_ssl_timeout(exc):
-                    self.logger.debug(
-                        "[%s(%s):%s] ssl timeout in : _read_bytes_from_socket",
-                        self.fn or "",
-                        self.host,
-                        self.port,
-                    )
-                    continue
-                raise
         return b"".join(chunks)
 
-    def _read_message(self):
+    def _read_message(self) -> CastMessage:
         """Reads a message from the socket and converts it to a message."""
         # first 4 bytes is Big-Endian payload length
         payload_info = self._read_bytes_from_socket(4)
@@ -848,22 +827,23 @@ class SocketClient(threading.Thread):
         # now read the payload
         payload = self._read_bytes_from_socket(read_len)
 
-        message = cast_channel_pb2.CastMessage()
+        message = CastMessage()
         message.ParseFromString(payload)
 
         return message
 
-    # pylint: disable=too-many-arguments
+    # pylint: disable=too-many-arguments, too-many-branches
     def send_message(
         self,
-        destination_id,
-        namespace,
-        data,
-        inc_session_id=False,
-        callback_function=False,
-        no_add_request_id=False,
-        force=False,
-    ):
+        destination_id: str,
+        namespace: str,
+        data: dict,
+        *,
+        inc_session_id: bool = False,
+        callback_function: CallbackType | None = None,
+        no_add_request_id: bool = False,
+        force: bool = False,
+    ) -> None:
         """Send a message to the Chromecast."""
 
         # namespace is a string containing namespace
@@ -873,7 +853,6 @@ class SocketClient(threading.Thread):
         # If channel is not open yet, connect to it.
         self._ensure_channel_connected(destination_id)
 
-        request_id = None
         if not no_add_request_id:
             request_id = self._gen_request_id()
             data[REQUEST_ID] = request_id
@@ -881,22 +860,20 @@ class SocketClient(threading.Thread):
         if inc_session_id:
             data[SESSION_ID] = self.session_id
 
-        msg = cast_channel_pb2.CastMessage()
+        msg = CastMessage()
 
-        msg.protocol_version = msg.CASTV2_1_0  # pylint: disable=no-member
+        msg.protocol_version = msg.CASTV2_1_0
         msg.source_id = self.source_id
         msg.destination_id = destination_id
-        msg.payload_type = (
-            cast_channel_pb2.CastMessage.STRING  # pylint: disable=no-member
-        )
+        msg.payload_type = CastMessage.STRING
         msg.namespace = namespace
-        msg.payload_utf8 = _json_to_payload(data)
+        msg.payload_utf8 = json.dumps(data, ensure_ascii=False)
 
         # prepend message with Big-Endian 4 byte payload size
         be_size = pack(">I", msg.ByteSize())
 
         # Log all messages except heartbeat
-        if msg.namespace != NS_HEARTBEAT:  # pylint: disable=no-member
+        if msg.namespace != NS_HEARTBEAT:
             self.logger.debug(
                 "[%s(%s):%s] Sending: %s",
                 self.fn or "",
@@ -906,17 +883,25 @@ class SocketClient(threading.Thread):
             )
 
         if not force and self.stop.is_set():
+            if callback_function:
+                callback_function(False, None)
             raise PyChromecastStopped("Socket client's thread is stopped.")
         if not self.connecting and not self._force_recon:
+            # We have a socket
+            assert self.socket is not None
+
             try:
                 if callback_function:
                     if not no_add_request_id:
                         self._request_callbacks[request_id] = callback_function
                     else:
-                        callback_function(None)
+                        callback_function(True, None)
                 self.socket.sendall(be_size + msg.SerializeToString())
             except socket.error:
-                self._request_callbacks.pop(request_id, None)
+                if callback_function:
+                    callback_function(False, None)
+                if not no_add_request_id:
+                    self._request_callbacks.pop(request_id, None)
                 self._force_recon = True
                 self.logger.info(
                     "[%s(%s):%s] Error writing to socket.",
@@ -925,57 +910,70 @@ class SocketClient(threading.Thread):
                     self.port,
                 )
         else:
+            if callback_function:
+                callback_function(False, None)
             raise NotConnected(f"Chromecast {self.host}:{self.port} is connecting...")
 
     def send_platform_message(
         self,
-        namespace,
-        message,
-        inc_session_id=False,
-        callback_function_param=False,
-        no_add_request_id=False,
-    ):
+        namespace: str,
+        message: dict,
+        *,
+        inc_session_id: bool = False,
+        callback_function: CallbackType | None = None,
+        no_add_request_id: bool = False,
+    ) -> None:
         """Helper method to send a message to the platform."""
         return self.send_message(
             PLATFORM_DESTINATION_ID,
             namespace,
             message,
-            inc_session_id,
-            callback_function_param,
+            inc_session_id=inc_session_id,
+            callback_function=callback_function,
             no_add_request_id=no_add_request_id,
         )
 
     def send_app_message(
         self,
-        namespace,
-        message,
-        inc_session_id=False,
-        callback_function_param=False,
-        no_add_request_id=False,
-    ):
+        namespace: str,
+        message: dict,
+        *,
+        inc_session_id: bool = False,
+        callback_function: CallbackType | None = None,
+        no_add_request_id: bool = False,
+    ) -> None:
         """Helper method to send a message to current running app."""
         if namespace not in self.app_namespaces:
+            if callback_function:
+                callback_function(False, None)
             raise UnsupportedNamespace(
                 f"Namespace {namespace} is not supported by current app. "
                 f"Supported are {', '.join(self.app_namespaces)}"
             )
 
+        if self.destination_id is None:
+            if callback_function:
+                callback_function(False, None)
+            raise NotConnected(
+                "Attempting send a message when destination_id is not set"
+            )
+
         return self.send_message(
             self.destination_id,
             namespace,
             message,
-            inc_session_id,
-            callback_function_param,
+            inc_session_id=inc_session_id,
+            callback_function=callback_function,
             no_add_request_id=no_add_request_id,
         )
 
-    def register_connection_listener(self, listener: ConnectionStatusListener):
+    def register_connection_listener(self, listener: ConnectionStatusListener) -> None:
         """Register a connection listener for when the socket connection
         changes. Listeners will be called with
         listener.new_connection_status(status)"""
         self._connection_listeners.append(listener)
 
-    def _ensure_channel_connected(self, destination_id):
+    def _ensure_channel_connected(self, destination_id: str) -> None:
         """Ensure we opened a channel to destination_id."""
         if destination_id not in self._open_channels:
             self._open_channels.append(destination_id)
@@ -999,7 +997,7 @@ class SocketClient(threading.Thread):
                 no_add_request_id=True,
             )
 
-    def disconnect_channel(self, destination_id):
+    def disconnect_channel(self, destination_id: str) -> None:
         """Disconnect a channel with destination_id."""
         if destination_id in self._open_channels:
             try:
@@ -1021,11 +1019,11 @@ class SocketClient(threading.Thread):
 
             self.handle_channel_disconnected()
 
-    def handle_channel_disconnected(self):
+    def handle_channel_disconnected(self) -> None:
         """Handles a channel being disconnected."""
         for namespace in self.app_namespaces:
             if namespace in self._handlers:
-                for handler in self._handlers[namespace]:
+                for handler in set(self._handlers[namespace]):
                     handler.channel_disconnected()
 
         self.app_namespaces = []
@@ -1036,15 +1034,18 @@ class SocketClient(threading.Thread):
 class ConnectionController(BaseController):
     """Controller to respond to connection messages."""
 
-    def __init__(self):
+    def __init__(self) -> None:
         super().__init__(NS_CONNECTION)
 
-    def receive_message(self, message, data: dict):
+    def receive_message(self, message: CastMessage, data: dict) -> bool:
         """
         Called when a message is received.
 
         data is message.payload_utf8 interpreted as a JSON dict.
         """
+        if self._socket_client is None:
+            raise ControllerNotRegistered
+
         if self._socket_client.is_stopped:
             return True
 
@@ -1063,17 +1064,20 @@ class ConnectionController(BaseController):
 class HeartbeatController(BaseController):
     """Controller to respond to heartbeat messages."""
 
-    def __init__(self):
+    def __init__(self) -> None:
         super().__init__(NS_HEARTBEAT, target_platform=True)
-        self.last_ping = 0
+        self.last_ping = 0.0
         self.last_pong = time.time()
 
-    def receive_message(self, _message, data: dict):
+    def receive_message(self, _message: CastMessage, data: dict) -> bool:
         """
         Called when a heartbeat message is received.
 
         data is message.payload_utf8 interpreted as a JSON dict.
         """
+        if self._socket_client is None:
+            raise ControllerNotRegistered
+
         if self._socket_client.is_stopped:
             return True
 
@@ -1099,8 +1103,11 @@ class HeartbeatController(BaseController):
 
         return False
 
-    def ping(self):
+    def ping(self) -> None:
         """Send a ping message."""
+        if self._socket_client is None:
+            raise ControllerNotRegistered
+
         self.last_ping = time.time()
         try:
             self.send_message({MESSAGE_TYPE: TYPE_PING})
@@ -1109,11 +1116,11 @@ class HeartbeatController(BaseController):
                 "Chromecast is disconnected. Cannot ping until reconnected."
             )
 
-    def reset(self):
+    def reset(self) -> None:
         """Reset expired counter."""
         self.last_pong = time.time()
 
-    def is_expired(self):
+    def is_expired(self) -> bool:
         """Indicates if connection has expired."""
         if time.time() - self.last_ping > HB_PING_TIME:
             self.ping()
@@ -1121,7 +1128,7 @@ class HeartbeatController(BaseController):
         return (time.time() - self.last_pong) > HB_PING_TIME + HB_PONG_TIME
 
 
-def new_socket():
+def new_socket() -> socket.socket:
     """
     Create a new socket with OS-specific parameters
 
